import sys
from pickle import NONE
from typing import List, Set
import importlib


def JsonSubTypes(subclassnames:List):
    '''
    A bit complex mechanism that just sets
    attribute __jsonsubtypes__ of the annotated class to the tuple
    (annotatedclass, subclassnames).
    We store annotatedclass itself also here because derived
    classes inherit this annotation, while they are NOT having
    these derived classes. We can check the annotated class to 
    check if this attribute holds.
    
    NOTE this annotation must be used only if the referred
    classes actually extend the annotated class.
    Failure to do this will lead to runtime errors.
    We follow the behaviour of jackson here:
    such classes may still serialize but will fail to deserialize.
    
    NOTE2 in jackson the parent-child relations are stored
    in a separate object. Here we store these in the 
    class object itself. The major advantage is that 
    we can leave inheritance to python inheritance mechanisms.
    Disadvantage is that this modifies the class,
    and that this can not handle multiple inheritance situations.
    Now I have no idea how these should be handled anyway,
    neither do I think jackson would support that as multiple
    inheritance is not part of java,
    so probably not much is currently lost here. 
    
    NOTE3 the list usually contains STRING objects with the
    full.qualified.path.to.Class. And Class is not "import"ed.
    This to avoid circular imports.
    
    @param subclassnames a list of full class names of subclasses.
    '''
    if not isinstance(subclassnames, list):
        raise ValueError("JSonSubTypes requires list of classes")
    # we can't check class existence here, as subclasses 
    # will not yet exist as they extend superclass.
    
    
    def doAnnotate(annotatedclass):
        '''
        This function is applied to the annotated class.
        It just adds an attribute __jsonsubtypes__ to the class
        with the annotation data
        '''
        setattr(annotatedclass,'__jsonsubtypes__',  subclassnames)
        return annotatedclass
    
    # we return doAnnotate as result of the function.
    # that way we can apply that function to the next argument
    # which is the annotatedclass
    return doAnnotate

def str_to_class(fullclasspath:str)->type:
    '''
    @param fullclasspath : full path specification to a class,
    so that we can locate and load it. 
    @return a new class, loaded from the given string.
    '''
    x=fullclasspath.rfind(".")
    module_name=fullclasspath
    if x>0:
        module_name=fullclasspath[0:x]
    # load the module, will raise ImportError if module cannot be loaded
    try:
        m = importlib.import_module(module_name)
    except Exception as e:
        raise ValueError("Failed to load class "+fullclasspath,e)
    # get the class, will raise AttributeError if class cannot be found
    c = getattr(m, fullclasspath[x+1:])
    return c

def getSubTypes(clas) -> Set:
    '''
    @param clas the class to get subtypes for.
    @return [subclasses] of clas, including clas.
    This is a list of Class objects
    of the @JsonSubTypes-registered subclasses of the annotated class.
    It also includes subclasses of the subclasses etc.
    The class itself is always member of the returned set.
    This annotation is inherited by subclasses, with 
    proper restriction to real sub classes,
    so if C has another class X annotated as subclass,
    but X is not actually a subclass of C, X is not added.
    It is assumed there are no cycles in the class hierarchy.
    '''
    subclassnames =  getattr(clas, '__jsonsubtypes__', [])
    subclasses = [ str_to_class(name) for name in subclassnames ]
    # filter the real subclasses
    subclasses = [ c for c in subclasses if issubclass(c, clas) and c!=clas ]
    allclasses=[clas]
    for clazz in subclasses:
        allclasses.extend(getSubTypes(clazz))
    return set(allclasses)


