from abc import ABCMeta
from datetime import datetime
from decimal import Decimal
import json
import time
from typing import cast , Dict, Any, List, Optional, TypeVar, Set,Collection 
from typing import TypeVar, Type
from uuid import UUID
import re


from pyson.JsonDeserialize import getDeserializer
from pyson.JsonGetter import getGetter
from pyson.JsonTools import isPrimitive, getActualClass, getInitArgs, getDefaults, addTypeInfo,\
    fullclasspath, str_to_class, uniforn_str_to_class
from pyson.JsonTypeInfo import Id, As, getTypeWrappingInfo
from pyson.JsonValue import getJsonValue
from uri.uri import URI
from pyson.JsonSerialize import getSerializer
from numbers import Number


class ObjectMapper:  
    '''
    A very simple pyson-style objectmapper.
    '''
    
    T=TypeVar("T")
    def parse(self, data:object, clas:T )->T:
        '''
        @param data either a dict or a built-in object like an int.
        @param clas the expected class contained in the data.
        If a dict, this class must have a __init__ function specifying the params
        needed and the data in this case should be a dict.
        Any is considered indicating that the expected argument is primitive.
        If this is a str, it is assumed to contain a string representation of a class,
        and is immediately evaluated to get its actual value.
        @return a clas instance matching the data
        '''
        if isinstance(clas, str):
            clas = eval(clas)
        
        # if there is deserializer, use it.
        deserializer=getDeserializer(clas)
        if deserializer:
            return deserializer().deserialize(data, clas) 

            
        
        # now distinguish 
        if isPrimitive(clas,True) or clas==Any:
            return self.parseBase(data, clas)
        if (clas==list):
            raise ValueError("Illegal type, use List[X] instead of list")
        if (clas==dict):
            raise ValueError("Illegal type, use Dict[X] instead of dict")
        
        
        if getTypeWrappingInfo(clas):
            if not isinstance(data, dict):
                raise ValueError("Expected class, therefore data must be a dict but got "+str(data))
            (data, clas) = getActualClass(cast(dict, data),clas)


        jsonvaluetype = self._getJsonValue(clas)
        if jsonvaluetype:
            return clas(self.parse(data, jsonvaluetype))

        # if getJsonValue(clas):
        #     argclasses:dict=getInitArgs(clas)
        #     if len(argclasses)!=1:
        #         raise ValueError("Class "+str(clas)+" has @JsonValue getter but has multiple init arguments")
        #     valueclas = list(argclasses.values())[0]
        #     return clas(self.parse(data, valueclas))

        if clas.__module__=='typing':
            return self.parseGeneric(data,clas)
        # after this, clas is NOT a generic like List[X] 
        if issubclass(clas, BaseException): #type:ignore
            return self.parseException(data,clas)
        
        if type(data)==dict: # if it contains class, data must be dict
            return self.parseClass(cast(dict,data), clas)
        


        raise ValueError("Expected "+str(clas)+" but got '"+str(data)+"' which is a "+str(type(data)))
        
 
    
    def parseBase(self,obj, clas:T)->T:
        '''
        @param obj a built-in object like an int
        @param clas the class of the expected object.
        If Any, then the obj is accepted as primitive (eg, int, dict, str)
        Else, if clas does not contain __init__, it must be a primitive
        @return the obj, after checking it's indeed a clas
        '''

        
        #datetime is special, the json primitive will be a int with millis since 1970
        if clas== datetime:
            if not type(obj)==int:
                raise ValueError("expected int (millis since 1970) but got "+str(obj))
            return datetime.fromtimestamp(cast(int, obj)/1000.0)
        if clas==URI:
            if not type(obj)==str:
                raise ValueError("expected uri string but got "+str(obj))
            return URI(obj)
        if clas==UUID:
            if not type(obj)==str:
                raise ValueError("expected UUID string but got "+str(obj))
            return UUID(obj)
        if isinstance(obj, Number) and type(obj)!=clas: #443 auto casting to required type
            if clas==float:
                return float(obj)
            if clas==int:
                return int(obj)
            if clas==Decimal:
                # We use str to get the number truncated at correct #digits.
                # It is possible json already rounded the number
                return Decimal(str(obj))
        if  not (clas==Any or type(obj)==clas):
            raise ValueError("Expected "+str(clas)+" but got "+str(obj)+" which is a "+str(type(obj)))
        return obj
             
 
    def parseClass(self, data:dict, clas:T)->T:
        '''
        @param data a dict with the values for class.__init__ function
        @return a clas instance matching the data
        '''
        if not isinstance(data,dict ):
            raise ValueError("data "+str(data)+" must be dict")
        # then this class needs initialization
        initargs:List=[]
        argclasses=getInitArgs(clas)
        defaults = getDefaults(clas)
        if not set(data.keys()).issubset(set(argclasses.keys())):
            raise ValueError("Constructor of "+str(clas)+" requires arguments "+str(list(argclasses.keys()))+" but got "+str(list(data.keys())))
        for arg in argclasses:
            if arg in data:
                try:
                    initargs.append (self.parse(data[arg], argclasses[arg]))
                except ValueError as e:
                    raise ValueError("Error parsing "+str(data[arg])+ " as "+str(argclasses[arg]),e ) from e
            else: # does constructor have a default for the missing avlue?
                if not arg in defaults: 
                    # fallback: if arg not default, then add None.
                    # note that default args are at the end, we don't need to fill up those.
                    initargs.append(None)
        return clas(*initargs)
    
    def parseGeneric(self, data, clas:T)->T:
        '''
        A generic is a class from typing, eg soemthing like typing Dict[str,str]. 
        @param data may be list or dict,  depending on the exact clas
        @param clas the target class of the expected result
        @return instance of the clas. Don't know how to write this for typing
        '''
        # repr starts with "typing.", which is dropped by [7:]
        gname =repr(clas)[7:] # _name fails for eg typing.Union
        
        if re.match("(List|Set|Collection).*", gname):
            elementclas = clas.__args__[0] if hasattr(clas,'__args__') else Any
            if type(data)!=list:
                raise ValueError("expected list[{elementclas}] but got "+str(data))
            if isinstance(elementclas , TypeVar):
                elementclas = Any # use primitive as fallback
            res=[self.parse(listitem, elementclas) for listitem in data]
            if gname.startswith('Set'):
                return set(res)
            return res
        
        if gname.startswith('Dict'):
            keyclas = clas.__args__[0] if hasattr(clas,'__args__') else Any
            if not keyclas.__hash__:
                raise ValueError("Dict cannot be deserialized, key class "+str(keyclas)+" does not  have __hash__")
            if isinstance(keyclas, TypeVar): # key type not specified
                keyclas=Any
            elementclas = clas.__args__[1] if hasattr(clas,'__args__') else Any
            if isinstance(elementclas, TypeVar): # value not specified
                elementclas=Any
            if type(data)!=dict:
                raise ValueError("expected dict[{keyclass, elementclas}] but got "+str(data))
            if keyclas==str or keyclas==Any or self._getJsonValue(keyclas):
                # the case where the target class is @JsonValue has been caught already.
                return { self.parse(key, keyclas) : self.parse(val, elementclas)\
                     for key,val in data.items() }
            else:
                # SPECIAL #190 in json keys are always string. Parse it as json if no string expected
                return { self.parse(json.loads(key), keyclas) : self.parse(val, elementclas)\
                     for key,val in data.items() }
 
            
        if gname.startswith('Union'):
            # Special to support optional: Union[class, NoneType]
            actualclasses = [actual for actual in clas.__args__ if actual!=type(None)]
            if len(actualclasses)!=1:
                raise ValueError("Union type only supported with NoneType, but found: "+str(clas))
            if data==None:
                return None
            return self.parse(data, actualclasses[0])
        
        if gname.startswith('Optional'):
            if data==None:
                return None
            return self.parse(data, clas.__args__[0])
        
        if gname.startswith('Type'):
            if not isinstance(data,str):
                raise ValueError("Expected string with full.class.path but got "+str(data))
            actual:Type = uniforn_str_to_class(data)
            basetype:Type = str_to_class(gname[4:].replace('[','').replace(']',''))
            if not issubclass(actual, basetype):
                raise ValueError("class "+str(data)+' is not instance of '+str(basetype) )
            return actual
        
        raise ValueError("Unsupported generic type "+ str(clas))
    
    def parseException(self, data, excclass=Exception):
        '''
        Assumes argument for excclass constructor is the message.
        '''
        if not isinstance(data, dict):
            raise ValueError("Expected dict  but got "+repr(data))
        if not "message" in data:
            raise ValueError("Expected 'message' in "+str(data))
        res=excclass(data['message'])
        if 'cause' in data:
            res.__cause__=self.parse(data['cause'], Optional[Exception])
        return res
    
    def toJson(self, data)->dict:
        '''
        @param data either a dict or a built-in object like an int
        @return a json structure containing this data 
        '''
        if isinstance(data, BaseException):
            return self.toJsonException(data)
        res:dict
        clas = type(data)
        # dict and list are handled separately because
        # we must serizliae keys and values separately
        if isPrimitive(clas):
            return self.toJsonBase(data) 
        if clas==list or clas==tuple or clas==set or clas==frozenset:
            res=self.toJsonList(data)
        elif clas==dict:
            res=self.toJsonDict(data)
        elif type(clas)==type or type(clas)==ABCMeta:
            # is it general class? FIXME can this be done better?
            res = self.toJsonClass(data)
        else:
            raise ValueError("Unsupported object of type "+str(clas))
        # check if wrapper setting is requested for this class
        if getTypeWrappingInfo(clas):
            res=addTypeInfo(clas,res);
        return res
        
    def toJsonClass(self,data:object)->object:
        ''' 
        @param data a class instance
        @return data a dict with the values 
        The values are based on the  class.__init__ function
        '''
        serializer = getSerializer(data)
        if serializer:
            return serializer().serialize(data)
        
        if isinstance(data, Type): # a class reference
            return fullclasspath(cast( Type,data))
            
        jsonvalue = getJsonValue(data)
        if jsonvalue:
            return self.toJson(jsonvalue())
        
        clas=type(data)
        res={}
        arg:str
        for arg in getInitArgs(clas):
            argvalue = getGetter(data, arg)()
#             gettername = 'get'+arg
#             if not hasattr(data, gettername) :# in clas.__dict__:
#                 raise ValueError("The object "+str(data)+ "of type "+ str(clas)+" has no function "+gettername)
#             argvalue = getattr(data, gettername)() #.__dict__[gettername]()
            res[arg]=self.toJson(argvalue)
        return res
        
    
    def toJsonBase(self,obj):
        '''
        @param obj a built-in object like an int or a datetime.
        obj does not contain __init__, it must be a built-in type
        @return the json style representation of teh built-in object.
        For datetime objects we use the linux timestamp (millis since 1970)
        '''
        if isinstance(obj, datetime):
            return round(datetime.timestamp(obj)*1000)
        if isinstance(obj, URI) or isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, Decimal):
            if obj == obj.to_integral_value():
                return int(obj) # python has no maxint 
            return float(obj)
        return obj
    
    def toJsonList(self, listofobj):
        '''
        @param listofobj list or tuple of objects each to be serialized separately.
        @return list object to be put in the json representation,  
        '''
        if len(listofobj)==0:
            return [] # empty list has no type. 
        #clas = getListClass(listofobj)
        #         if isPrimitive(clas):
        #             return listofobj
        # CHECK can we check if classes involved are properly annotated?
        #if not(clas==None or getTypeWrappingInfo(clas)):
        #    raise ValueError("@JsonTypeInfo is required for list objects, but found "+str(clas))
        return [self.toJson(elt) for elt in listofobj]
    
    def toJsonDict(self, dictofobj:Dict[Any,Any]):
        '''
        @param dictofobj dict with objects each to be serialized separately.
        The keys must be primitive, values must be all the same class.
        @return list object to be put in the json representation,  
        if keys are not primitive objects, they are turned into 
        strings because json only allows string objects as key.
        '''
        return { self.keytojson(key):self.toJson(val) for key,val in dictofobj.items()}
    
    def keytojson(self,key):
        '''
        @param key an object to turn into json
        @return a json-version of the key object. If the key is not a primitive object
        (int, str, ...) then the json value is turned into a string
        '''
        if (type(key)==str):
            return key
        if (self._getJsonValue(key)==str):
            # special case: object is @JsonValue of type str. 
            return self.toJson(key)
        return json.dumps(self.toJson(key))

    def _getJsonValue(self, clas):
        '''
        @param clas the class to check.
        @return The class of the argument type needed for constructing an instance of this class,
        or None if clas is not annotated with @JsonValue. 
        '''
        if not getJsonValue(clas):
            return None
        argclasses:dict=getInitArgs(clas)
        if len(argclasses)!=1:
            raise ValueError("Class "+str(clas)+" has @JsonValue getter, but has "+str(len(argclasses))+" init arguments")
        return list(argclasses.values())[0]

       
    def toJsonException(self, e:BaseException):
        res:Dict[str ,Any] ={}
        if len(e.args)==1:
            res["message"]=str(e.args[0])
        else:
            res["message"]=str(e.args)
        res["stackTrace"]=[]
        res["cause"]=self.toJsonException(e.__cause__) if e.__cause__ else None
        return res
    